From de5aa1305b60d81425a481346f6c99e83fe909d1 Mon Sep 17 00:00:00 2001
From: Shugo Maeda <shugo@ruby-lang.org>
Date: Sat, 30 Oct 2010 01:55:51 +0900
Subject: [PATCH 40/51] do not overlay the same module twice.
 do not use omod to avoid infinite loop in invokesuper.

---
 eval.c      |    7 +++++++
 insns.def   |    3 +--
 method.h    |    1 +
 vm_method.c |   22 +++++++++++++++-------
 4 files changed, 24 insertions(+), 9 deletions(-)

diff --git a/eval.c b/eval.c
index ede63ff..f69444a 100644
--- a/eval.c
+++ b/eval.c
@@ -877,6 +877,13 @@ rb_overlay_module(NODE *cref, VALUE klass, VALUE module)
 	}
 	if (!NIL_P(c = rb_hash_lookup(cref->nd_omod, klass))) {
 	    superclass = c;
+	    while (c && TYPE(c) == T_ICLASS) {
+		if (RBASIC(c)->klass == module) {
+		    /* already overlayed module */
+		    return;
+		}
+		c = RCLASS_SUPER(c);
+	    }
 	}
     }
     FL_SET(module, RMODULE_IS_OVERLAYED);
diff --git a/insns.def b/insns.def
index 3e3b808..98300c1 100644
--- a/insns.def
+++ b/insns.def
@@ -1064,12 +1064,11 @@ invokesuper
     while (ip && !ip->klass) {
 	ip = ip->parent_iseq;
     }
-  again:
     me = rb_method_entry(klass, id, &klass);
     if (me && me->def->type == VM_METHOD_TYPE_ISEQ &&
 	me->def->body.iseq == ip) {
 	klass = RCLASS_SUPER(klass);
-	goto again;
+	me = rb_method_entry_get_with_omod(Qnil, klass, id, &klass);
     }
 
     CALL_METHOD(num, blockptr, flag, id, me, recv, klass);
diff --git a/method.h b/method.h
index b0a92e8..4750b45 100644
--- a/method.h
+++ b/method.h
@@ -91,6 +91,7 @@ void rb_add_method_cfunc(VALUE klass, ID mid, VALUE (*func)(ANYARGS), int argc,
 rb_method_entry_t *rb_add_method(VALUE klass, ID mid, rb_method_type_t type, void *option, rb_method_flag_t noex);
 rb_method_entry_t *rb_method_entry(VALUE klass, ID id, VALUE *define_class_ptr);
 
+rb_method_entry_t *rb_method_entry_get_with_omod(VALUE omod, VALUE klass, ID id, VALUE *define_class_ptr);
 rb_method_entry_t *rb_method_entry_get_without_cache(VALUE klass, VALUE omod, ID id, VALUE *define_class_ptr);
 rb_method_entry_t *rb_method_entry_set(VALUE klass, ID mid, const rb_method_entry_t *, rb_method_flag_t noex);
 
diff --git a/vm_method.c b/vm_method.c
index d5ebf75..2d362f7 100644
--- a/vm_method.c
+++ b/vm_method.c
@@ -461,15 +461,10 @@ rb_method_entry_get_without_cache(VALUE klass, VALUE omod, ID id, VALUE *defined
 }
 
 rb_method_entry_t *
-rb_method_entry(VALUE klass, ID id, VALUE *defined_class_ptr)
+rb_method_entry_get_with_omod(VALUE omod, VALUE klass, ID id,
+			      VALUE *defined_class_ptr)
 {
     struct cache_entry *ent;
-    NODE *cref = rb_vm_cref();
-    VALUE omod = Qnil;
-
-    if (cref && !NIL_P(cref->nd_omod)) {
-	omod = cref->nd_omod;
-    }
 
     ent = cache + EXPR1(klass, omod, id);
     if (ent->mid == id && ent->klass == klass && ent->omod == omod) {
@@ -482,6 +477,19 @@ rb_method_entry(VALUE klass, ID id, VALUE *defined_class_ptr)
 					     defined_class_ptr);
 }
 
+rb_method_entry_t *
+rb_method_entry(VALUE klass, ID id, VALUE *defined_class_ptr)
+{
+    struct cache_entry *ent;
+    NODE *cref = rb_vm_cref();
+    VALUE omod = Qnil;
+
+    if (cref && !NIL_P(cref->nd_omod)) {
+	omod = cref->nd_omod;
+    }
+    return rb_method_entry_get_with_omod(omod, klass, id, defined_class_ptr);
+}
+
 static void
 remove_method(VALUE klass, ID mid)
 {
-- 
1.7.0.4

